/*
 * Adi Fairbank
 * CS 490, Spring 1998
 * Instructor: Bruce Land
 * Project: Cryptography, Computer Security
 *
 * Description of this program: Implementation of RSA algorithm
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX 256
#define TRUE 1
#define FALSE 0
#define LEFT 1
#define RIGHT 0

typedef struct {
  unsigned char e[MAX];
  unsigned char n[MAX];
} RSA_public_key;

typedef struct {
  unsigned char d[MAX];
  unsigned char n[MAX];
} RSA_private_key;

typedef short int boolean;


/*
 * show_value()
 */
void show_value(const unsigned char *val, const unsigned int digits)
{
  int i;

  for (i = MAX - digits; i < MAX; i++)
    printf("%.2x.", val[i]);
}


/*
 * valzero()
 *
 * desc: returns TRUE if the given MAX-byte number is zero
 *       returns FALSE otherwise
 */
boolean valzero(const unsigned char *a)
{
  int i;

  for (i = 0; i < MAX; i++)
    if (a[i] != 0)
      return FALSE;

  return TRUE;
}


/*
 * valcmp()
 *
 * desc: an all-purpose value comparer; compares two MAX-byte numbers.
 *       returns 0  if a == b
 *       returns 1  if a > b
 *       returns -1 if a < b
 */
char valcmp(const unsigned char *a,
	    const unsigned char *b)
{
  int i;

  for (i = 0; i < MAX; i++) {
    if (a[i] > b[i])
      return 1;
    if (a[i] < b[i])
      return -1;
  }

  return 0;
}


/*
 * valcpy()
 *
 * desc: copies all bits of the MAX-byte number 'src' to the
 *       MAX-byte number 'dest'
 */
void valcpy(const unsigned char *src,
	    unsigned char *dest)
{
  int i;

  for (i = 0; i < MAX; i++)
    dest[i] = src[i];
}


/*
 * vallen()
 *
 * desc: returns the number of significant bytes in the given MAX-byte number
 */
unsigned int vallen(const unsigned char *a)
{
  unsigned int i;

  for (i = 0; i < MAX; i++)
    if (a[i] != 0)
      return (MAX - i);

  return 0;
}


/*
 * addchar()
 *
 * desc: adds two unsigned bytes (each value ranges from 0 - 255) with a carry bit
 *       and returns the result and a carry out bit.
 */
void addchar(const unsigned char a,
	     const unsigned char b,
	     const unsigned char carry_in,
	     unsigned char *result,
	     unsigned char *carry_out)
{
  *result = a + b + carry_in;
  if (a + b + carry_in > 255)
    *carry_out = 1;
  else
    *carry_out = 0;
}


/*
 * subchar()
 *
 * desc: subtracts two unsigned bytes (each value ranges from 0 - 255) with a carry bit
 *       and returns the result plus the carry outbit.
 */
void subchar(const unsigned char a,
	     const unsigned char b,
	     const unsigned char carry_in,
	     unsigned char *result,
	     unsigned char *carry_out)
{
  unsigned char tmp = a;
  *result = 0;
  *carry_out = 0;

  /*
   * special case, if a == 0 we must make sure not to subtract the carry_in
   *   from it which would create a value of 255 (there would never be a
   *   carry_out either, so it would be cheating)
   */
  if (a == 0) {
    if (carry_in || b > 0) {
      *carry_out = 1;
      *result = 256 - carry_in - b;
    }
    return;
  }

  /*
   * first subtract the carry_in, then calculate the result and carry_out
   */
  tmp -= carry_in;
  if (tmp < b) {
    *carry_out = 1;
    *result = tmp + 256 - b;
  } else {
    *carry_out = 0;
    *result = tmp - b;
  }
}


/*
 * add()
 *
 * desc: adds two MAX-byte numbers and stores the result in 'result'
 *       accomplishes this by chaining addchar() routines together with carries
 */
unsigned char *add(const unsigned char *a,
		   const unsigned char *b,
		   unsigned char *result)
{
  unsigned int i;
  unsigned char carry = 0;

  for (i = MAX - 1; i > 0; i--)
    addchar(a[i], b[i], carry, &result[i], &carry);

  return result;
}


/*
 * sub()
 *
 * desc: subtracts two MAX-byte numbers and stores the result in 'result'
 *       accomplishes this by chaining subchar() routines together with borrows
 */
unsigned char *sub(const unsigned char *a,
		   const unsigned char *b,
		   unsigned char *result)
{
  unsigned int i;
  unsigned char borrow = 0;

  for (i = MAX - 1; i > 0; i--)
    subchar(a[i], b[i], borrow, &result[i], &borrow);

  return result;
}


/*
 * decrem()
 *
 * desc: decrements the given MAX-byte number
 *       does not iterate over all MAX bits
 */
unsigned char *decrem(unsigned char *a)
{
  int i = MAX - 1;
  unsigned char borrow = 1;

  while (borrow && i >= 0) {
    subchar(a[i], 0, borrow, &a[i], &borrow);
    i--;
  }

  return a;
}


/*
 * increm()
 *
 * desc: increments the given MAX-byte number
 *       does not iterate over all MAX bits
 */
unsigned char *increm(unsigned char *a)
{
  int i = MAX - 1;
  unsigned char carry = 1;

  while (carry && i >= 0) {
    addchar(a[i], 0, carry, &a[i], &carry);
    i--;
  }

  return a;
}


/*
 * shift()
 *
 * desc: shifts a MAX-byte integer by 1 bit in the direction specified
 *       direction == 0  right shift
 *       direction == 1  left shift
 */
unsigned char *shift(unsigned char *a,
		     unsigned char direction)
{
  int i, carry_in = 0, carry_out = 0;

  if (direction)
    for (i = MAX - 1; i > 0; i--) {
      carry_out = (a[i] & 0x80) >> 7;
      a[i] <<= 1;
      a[i] |= carry_in;
      carry_in = carry_out;
    }
  else
    for (i = 0; i < MAX; i++) {
      carry_out = a[i] & 0x01;
      a[i] >>= 1;
      a[i] |= (carry_in << 7);
      carry_in = carry_out;

      /*      printf("SHIFT:  (right)  a[%d] = %d.%d.%d\n", i, a[1021], a[1022], a[1023]);*/
    }

  return a;
}


/*
 * mult()
 *
 * desc: multiplies two MAX-byte numbers and returns the result
 */
unsigned char *mult(const unsigned char *a,
		    const unsigned char *b,
		    unsigned char *result)
{
  unsigned char tmp_a[MAX];
  unsigned char tmp_b[MAX];
  unsigned char mult_count[MAX];
  unsigned char tmp_result[MAX];
  int i;

  for (i = 0; i < MAX; i++)
    tmp_result[i] = 0;

  valcpy(b, tmp_b);

  while (!valzero(tmp_b)) {
    /*
     * initial values,
     *   mult_count = 1
     *   tmp_a = a
     */
    for (i = 0; i < MAX; i++)
      mult_count[i] = 0;
    mult_count[MAX - 1] = 1;
    valcpy(a, tmp_a);

    /*
     * keep multiplying tmp_a by 2 until the total number of a's
     *   we have summed is greater than b
     */
    while (valcmp(mult_count, tmp_b) == -1) {
      shift(tmp_a, LEFT);
      shift(mult_count, LEFT);
    }

    if (valcmp(mult_count, tmp_b) == 1) {
      shift(tmp_a, RIGHT);
      shift(mult_count, RIGHT);
    }

    add(tmp_result, tmp_a, tmp_result);

    sub(tmp_b, mult_count, tmp_b);
  }

  /* old, slow way
   *
  while (!valzero(tmp_b)) {
    add(tmp_a, a, tmp_result);
    valcpy(tmp_result, tmp_a);
    decrem(tmp_b);
  }
  */

  valcpy(tmp_result, result);
  return result;
}


/*
 * expon()
 *
 * desc: computes b raised to the power e, both b and e being MAX-byte numbers
 */
unsigned char *expon(const unsigned char *b,
		     const unsigned char *e,
		     unsigned char *result)
{
  unsigned char tmp_b[MAX];
  unsigned char tmp_e[MAX];
  unsigned char exp_count[MAX];
  unsigned char prev[MAX];
  unsigned char tmp_result[MAX];
  int i;

  for (i = 0; i < MAX; i++)
    tmp_result[i] = 0;
  tmp_result[MAX - 1] = 1;

  valcpy(e, tmp_e);

  while (!valzero(tmp_e)) {

    /*
     * initial values,
     *   exp_count = 1
     *   tmp_b = b
     */
    for (i = 0; i < MAX; i++)
      exp_count[i] = 0;
    exp_count[MAX - 1] = 1;
    valcpy(b, tmp_b);

    /*
     * keep squaring tmp_b until we've got an exponent greater or equal
     * to tmp_e
     */
    while (valcmp(exp_count, tmp_e) == -1) {
      valcpy(tmp_b, prev);
      mult(tmp_b, tmp_b, tmp_b);
      shift(exp_count, LEFT);
    }

    if (valcmp(exp_count, tmp_e) == 1) {
      valcpy(prev, tmp_b);
      shift(exp_count, RIGHT);
    }

    mult(tmp_result, tmp_b, tmp_result);

    sub(tmp_e, exp_count, tmp_e);

    /*    printf("EXPON: tmp_result = ");
    show_value(tmp_result, 10);
    printf("\n\ttmp_e = ");
    show_value(tmp_e, 10);
    printf("\n\ttmp_b = ");
    show_value(tmp_b, 10);
    printf("\n");*/
  }

  /* old, slow way
   *
  while (!valzero(tmp_e)) {
    mult(tmp_b, b, tmp_result);
    valcpy(tmp_result, tmp_b);
    decrem(tmp_e);
  }
  */

  valcpy(tmp_result, result);
  return result;
}


/*
 * divide()
 *
 * desc: divides two MAX-byte numbers, stores the quotient in 'quotient'
 *                                and stores the remainder in 'remainder'
 */
unsigned char *divide(const unsigned char *numer,
		      const unsigned char *denom,
		      unsigned char *quotient,
		      unsigned char *remainder)
{
  unsigned char tmp_add[MAX];
  unsigned char tmp_quot[MAX];
  unsigned char tmp_sum[MAX];
  unsigned char final_quotient[MAX];
  unsigned char final_remainder[MAX];
  int i;

  /*
   * initial value,
   *   tmp_quot = 1
   */
  for (i = 0; i < MAX; i++) {
    final_quotient[i] = 0;
    tmp_quot[i] = 0;
    tmp_sum[i] = 0;
  }
  tmp_quot[MAX - 1] = 1;

  /*
   * initial values,
   *   tmp_add = denom   (one unit of denominator)
   *   final_remainder = numer   (the numerator)
   */
  valcpy(denom, tmp_add);
  valcpy(numer, final_remainder);

  /*  printf("### final_remainder = ");
  show_value(final_remainder, 10);
  printf("\n");
  printf("### done initial valcpys\n");*/

  /*
   * while tmp_sum < numer, multiply tmp_sum by two
   *
   *   algorithm: Keep multiplying denominator by two until it is bigger than
   *              numerator while keeping track of the number of times we did
   *              this.  When the value is bigger than the numerator, start
   *              back at 1, keep multiplying by two, and adding this value to
   *              the old.
   */
  while (valcmp(denom, final_remainder) != 1) {

    /*    printf("### shifting\n");*/
    while (valcmp(tmp_add, final_remainder) == -1) {

      shift(tmp_add, LEFT);
      shift(tmp_quot, LEFT);

      /*      printf("DIVIDE: tmp_add = ");
      show_value(tmp_add, 20);
      printf("\n\ttmp_quot = ");
      show_value(tmp_quot, 20);
      printf("\n");*/

    }

    if (valcmp(tmp_add, final_remainder) == 1) {
      shift(tmp_add, RIGHT);
      shift(tmp_quot, RIGHT);
    }

    /*    printf("### done shifting\n");*/

    add(tmp_sum, tmp_add, tmp_sum);
    add(final_quotient, tmp_quot, final_quotient);
    sub(numer, tmp_sum, final_remainder);

    valcpy(denom, tmp_add);
    for (i = 0; i < MAX; i++)
      tmp_quot[i] = 0;
    tmp_quot[MAX - 1] = 1;

    /*    printf("DIVIDE: tmp_sum = ");
    show_value(tmp_sum, 20);
    printf("\n\tfinal_quotient = ");
    show_value(final_quotient, 20);
    printf("\n\tfinal_remainder = ");
    show_value(final_remainder, 20);
    printf("\n");*/

  }

  /*  if (valcmp(tmp_sum, numer) == 1)
    sub(tmp_sum, denom, tmp_sum);

  sub(numer, tmp_sum, final_remainder);*/

  valcpy(final_quotient, quotient);
  valcpy(final_remainder, remainder);
  return quotient;
}


/*
 * sqroot()
 *
 * desc: computes the square root of a number 'x' by Newton's algorithm and
 *       stores the result in 'root'
 */
unsigned char *sqroot(const unsigned char *x,
		      unsigned char *root)
{
  unsigned char y[MAX];
  unsigned char y_prev[MAX];
  unsigned char tmp1[MAX];
  unsigned char tmp2[MAX];
  unsigned char one[MAX];
  int i;

  for (i = 0; i < MAX; i++) {
    y[i] = 0;
    y_prev[i] = 0;
    tmp1[i] = 0;
    tmp2[i] = 0;
    one[i] = 0;
  }
  y[MAX - 1] = 1;
  one[MAX - 1] = 1;

  do {
    valcpy(y, y_prev);
    divide(x, y, tmp1, tmp2);
    /*    printf("SQROOT: done divide\n");*/
    add(y, tmp1, tmp2);
    shift(tmp2, RIGHT);
    /*    printf("SQROOT: done shift\n");*/
    valcpy(tmp2, y);

  } while (valcmp(y, y_prev) == -1 || valcmp(y_prev, one) == 0);

  if (valcmp(y, y_prev) == 1)
    valcpy(y_prev, y);

  valcpy(y, root);

  return root;
}


/*
 * check_prime()
 *
 * desc: returns TRUE if large_integer is prime
 *       returns FALSE otherwise
 */
boolean check_prime(const unsigned char *large_integer)
{
  unsigned char stop[MAX];
  unsigned char divisor[MAX];
  unsigned char quot[MAX];
  unsigned char rem[MAX];
  /*  boolean recurse = FALSE;*/
  int i;

  /*
   * check if large_integer is even
   */
  if (large_integer[MAX - 1] & 0x01 == 0)
    return FALSE;

  /*
   * check if large_integer is divisible by 5
   */
  for (i = 0; i < MAX; i++)
    divisor[i] = 0;
  divisor[MAX - 1] = 5;

  /*  printf("### dividing by 5\n");*/
  divide(large_integer, divisor, quot, rem);
  if (valzero(rem))
    return FALSE;

  /*
   * start checking at 3
   */
  for (i = 0; i < MAX; i++)
    divisor[i] = 0;
  divisor[MAX - 1] = 3;

  /*
   * compute square root of large_integer
   */
  sqroot(large_integer, stop);

  /*
   * check numbers ending in 1,3,7,9 up to sqrt(large_integer) for divisibility
   */
  i = 3;
  while (valcmp(divisor, stop) != 1) {
    divide(large_integer, divisor, quot, rem);
    if (valzero(rem))
      return FALSE;

    /*
    if (recursion_level == 1) {
      divide(large_integer, divisor, quot, rem);
      if (valzero(rem))
	return FALSE;
    } else if (recurse)
      if (check_prime(divisor, recursion_level + 1)) {
	divide(large_integer, divisor, quot, rem);
	if (valzero(rem))
	  return FALSE;
      }

    if (!recurse && i == 9)
      recurse = TRUE;
      */

    increm(divisor);
    increm(divisor);
    i = (i + 2) % 10;

    if (i == 5) {
      increm(divisor);
      increm(divisor);
      i = 7;
    }
  }

  return TRUE;
}


/*
 * find_prime()
 *
 * desc: finds the smallest prime number greater than or equal to reference
 */
unsigned char *find_prime(const unsigned char *reference,
			  unsigned char *prime)
{
  int i;

  valcpy(reference, prime);

  /*
   * start at smallest odd number greater than reference point
   */
  prime[MAX - 1] |= 0x01;

  /*
   * continue checking odd numbers until a prime is found
   */
  i = prime[MAX - 1] % 10;
  /*  printf("### checking prime\n");*/
  while (!check_prime(prime)) {
    /*    printf("checked ");
    show_value(prime, 20);
    printf("\n");*/

    increm(prime);
    increm(prime);
    i = (i + 2) % 10;
    if (i == 5) {
      increm(prime);
      increm(prime);
      i = 7;
    }
  }

  return prime;
}


/*
 * generate_key_pair()
 *
 * desc: given two primes p and q, generates a full RSA key pair
 *       (public consists of (e, n) and private consists of (d, n))
 */
void generate_key_pair(const unsigned char *p,
		       const unsigned char *q,
		       RSA_public_key *pub,
		       RSA_private_key *priv)
{
  unsigned char n[MAX];
  unsigned char x0[MAX];
  unsigned char x1[MAX];
  unsigned char x2[MAX];
  unsigned char a[MAX];
  unsigned char b[MAX];
  unsigned char phi[MAX];
  unsigned char tmp[MAX];
  unsigned char tmp2[MAX];
  int i;

  /*
   * n = p * q
   */
  mult(p, q, n);

  /*
   * phi(n) = (p - 1)*(q - 1) = n - (p + q) + 1
   */
  add(p, q, phi);
  sub(n, phi, phi);
  increm(phi);

  /*
   * compute d as the smallest prime number greater than max(p*2, q*2)
   */
  valcpy(p, a);
  valcpy(q, b);
  shift(a, LEFT);
  shift(b, LEFT);
  if (valcmp(a, b) == 1)
    find_prime(a, priv->d);
  else
    find_prime(b, priv->d);

  /*  printf("GENRSA: d = ");
  show_value(priv->d, 20);
  printf("\n");*/

  /*
   * x0 = 0
   */
  for (i = 0; i < MAX; i++) {
    x0[i] = 0;
  }
  valcpy(priv->d, x1);
  divide(phi, x1, tmp, x2);

  /*
   * find e from d and phi(n)
   */

  /*
   * compute the x[i] values by x[i+1] = x[i-1] (mod x[i])
   * until we reach an x[i+1] = 0
   */
  while (!valzero(x2)) {
    valcpy(x1, x0);
    valcpy(x2, x1);
    divide(x0, x1, tmp, x2);

    /*    printf("GENRSA: x[i] = ");
    show_value(x2, 20);
    printf("\n");*/
  }

  /*
   * quick algorithm to find a, b: keep adding units of phi and d
   * to a and b, respectively until the difference between a and b
   * is equal to x1 (last nonzero x[i])
   */
  valcpy(phi, a);
  valcpy(priv->d, b);
  divide(a, b, tmp2, tmp);
  mult(tmp2, priv->d, tmp2);
  do {
    if (valcmp(a, b) != -1) {
      add(b, tmp2, b);
      while (valcmp(a, b) != -1)
	add(b, priv->d, b);
    } else {
      sub(b, a, tmp);
      add(a, phi, a);
    }
  } while (valcmp(tmp, x1) != 0);

  sub(a, phi, a);
  divide(b, priv->d, b, tmp);

  /*
   * add the last nonzero x[i] to phi and check for divisibility by priv->d
   * if we find one that is divisible, we know b
   * /
  valcpy(x1, tmp);
  while (!valzero(x0)) {
    /*    printf("GENRSA: tmp = ");
    show_value(tmp, 20);
    printf("\n");* /
    add(tmp, phi, tmp);
    divide(tmp, priv->d, b, x0);
  }
  */
  /*  printf("GENRSA: found b = ");
  show_value(b, 20);
  printf("\n");*/

  /*
   * subtract the last nonzero x[i] from the sum and compute a by
   * dividing the sum by phi
   * /
  sub(tmp, x1, tmp);
  divide(tmp, phi, a, x0);
  */

  /*  printf("GENRSA: found a = ");
  show_value(a, 20);
  printf("\n");*/

  /*
   * output the public encryption exponent
   */
  valcpy(b, pub->e);

  /*
   * output the value n
   */
  valcpy(n, priv->n);
  valcpy(n, pub->n);
}


/*
 * rsa_encrypt()
 *
 * desc: uses the given public key to RSA encrypt a
 *       MAX-byte block of plaintext
 */
unsigned char *rsa_encrypt(const RSA_public_key *key,
			   const unsigned char *plaintext,
			   unsigned char *ciphertext)
{
  unsigned char tmp[MAX];

  expon(plaintext, key->e, tmp);
  divide(tmp, key->n, tmp, ciphertext);
}


/*
 * rsa_encrypt_bitwise()
 *
 * desc:
 */
unsigned char *rsa_encrypt_bitwise(const RSA_public_key *key,
				   const unsigned char *plaintext,
				   unsigned char *ciphertext)
{
  unsigned char tmp[MAX];
  unsigned char mask;
  int i;

  /*
   * initial value,
   *   ciphertext = 1
   */
  for (i = 0; i < MAX; i++)
    ciphertext[i] = 0;
  ciphertext[MAX - 1] = 1;

  /*  printf("RSA_ENC: vallen(key->e) = %u\n", vallen(key->e));*/

  /*
   * encrypt using the algorithm given in section VII.A. of the RSA paper
   */
  for (i = vallen(key->e); i > 0; i--) {

    for (mask = 0x80; mask >= 0x01; mask >>= 1) {

      /*      printf("RSA_ENC: ciphertext = ");
      show_value(ciphertext, 20);
      printf("\n\ti = %d\n", i);
      printf("\tn = ");
      show_value(key->n, 20);
      printf("\n");*/

      mult(ciphertext, ciphertext, ciphertext);
      divide(ciphertext, key->n, tmp, ciphertext);
      if (key->e[MAX - i] & mask) {
	mult(ciphertext, plaintext, ciphertext);

	/*	printf("RSA_ENC: done mult, ciphertext =\n");
	show_value(ciphertext, 30);
	printf("\n");*/

	divide(ciphertext, key->n, tmp, ciphertext);

	/*	printf("RSA_ENC: done divide, ciphertext =\n");
	show_value(ciphertext, 30);
	printf("\n");*/
      }
    }

  }

  /*  printf("RSA_ENC: FINAL CIPHERTEXT\n\t");
  show_value(ciphertext, 30);
  printf("\n");*/
}


/*
 * rsa_decrypt()
 *
 * desc: uses the given private key to RSA decrypt a
 *       MAX-byte block of ciphertext
 */
void rsa_decrypt (const RSA_private_key *key,
		  const unsigned char *ciphertext,
		  unsigned char *plaintext)
{
  unsigned char tmp[MAX];

  expon(ciphertext, key->d, tmp);
  divide(tmp, key->n, tmp, plaintext);
}


/*
 * rsa_decrypt_bitwise()
 *
 * desc:
 */
unsigned char *rsa_decrypt_bitwise(const RSA_private_key *key,
				   const unsigned char *ciphertext,
				   unsigned char *plaintext)
{
  unsigned char tmp[MAX];
  unsigned char mask;
  int i;

  /*
   * initial value,
   *   plaintext = 1
   */
  for (i = 0; i < MAX; i++)
    plaintext[i] = 0;
  plaintext[MAX - 1] = 1;

  /*
   * decrypt using the algorithm given in section VII.A. of the RSA paper
   */
  for (i = vallen(key->d); i > 0; i--) {

    for (mask = 0x80; mask >= 0x01; mask >>= 1) {

      mult(plaintext, plaintext, plaintext);
      divide(plaintext, key->n, tmp, plaintext);

      if (key->d[MAX - i] & mask) {
	mult(plaintext, ciphertext, plaintext);
	divide(plaintext, key->n, tmp, plaintext);
      }

    }

  }
}


/*
 * main()
 *
 * desc: asks user for length of RSA key, and some random keystrokes
 *       then generates the key given that information
 */
void main()
{
  unsigned char a = 128;
  unsigned char b = 127;
  unsigned char c_in = 1, c_out;
  unsigned char d;
  unsigned char big1[MAX];
  unsigned char big2[MAX];
  unsigned char big3[MAX];
  unsigned char big4[MAX];
  unsigned char big5[MAX];
  unsigned char big6[MAX];
  unsigned char big7[MAX];
  unsigned char ranchar[MAX];
  unsigned char p[MAX];
  unsigned char q[MAX];
  unsigned char p_base[MAX];
  unsigned char q_base[MAX];
  unsigned char plaintext[MAX];
  unsigned char ciphertext[MAX];
  unsigned char tmp[MAX];
  unsigned int keylen;
  RSA_public_key *pubkey = malloc(sizeof(RSA_public_key));
  RSA_private_key *privkey = malloc(sizeof(RSA_private_key));
  int i;

  /*  addchar(a, b, c_in, &d, &c_out);
  printf("%d + %d + %d = %d  carryout = %d\n", a, b, c_in, d, c_out);*/

  for (i = 0; i < MAX; i++) {
    big1[i] = 0;
    big2[i] = 0;
    big3[i] = 0;
    big4[i] = 0;
    big5[i] = 0;
    big6[i] = 0;
    big7[i] = 0;
    ranchar[i] = 0;
    tmp[i] = 0;
    plaintext[i] = 0;
  }

  big1[MAX - 1] = 47;
  big1[MAX - 2] = 0;
  big1[MAX - 3] = 0;
  big2[MAX - 1] = 59;
  big2[MAX - 2] = 0;
  big4[MAX - 1] = 157;
  big5[MAX - 1] = 17;
  big6[MAX - 1] = 9;
  big7[MAX - 1] = 2;
  big7[MAX - 2] = 0;
  big7[MAX - 3] = 0;
  big7[MAX - 4] = 0;
  big7[MAX - 5] = 0;

  /*
  mult(big1, big2, big3);
  show_value(big1, 10);
  printf(" * ");
  show_value(big2, 10);
  printf("\n = ");
  show_value(big3, 10);
  printf("\n");
  */
  /*
  expon(big2, big4, big3);
  show_value(big2, 10);
  printf(" ^ ");
  show_value(big4, 10);
  printf("\n = ");
  show_value(big3, 100);
  printf("\n");

  divide(big3, big2, big4, big5);
  show_value(big3, 100);
  printf("\n / ");
  show_value(big2, 10);
  printf("\n = ");
  show_value(big4, 100);
  printf("\n remainder ");
  show_value(big5, 10);
  printf("\n");
  */

  /*  generate_key_pair(big1, big2, pubkey, privkey);
  valcpy(big4, privkey->d);
  valcpy(big5, pubkey->e);*/

  printf("Enter key length in bytes: ");
  scanf("%d", &keylen);

  printf("Please enter %d random characters on your keyboard: ", keylen << 1);
  scanf("%s", ranchar);
  for (i = 0; i < keylen; i++)
    p_base[MAX - i - 1] = ranchar[i];
  for (i = keylen; i < keylen << 1; i++)
    q_base[MAX - i + keylen - 1] = ranchar[i];

  printf("### finding primes p and q\n");

  find_prime(p_base, p);
  find_prime(q_base, q);

  printf("### generating key pair\n");

  generate_key_pair(p, q, pubkey, privkey);

  printf("p =\t");
  show_value(p, 20);
  printf("\nq =\t");
  show_value(q, 20);
  printf("\npublic key exponent e =\n\t");
  show_value(pubkey->e, 20);
  printf("\nprivate key exponent d =\n\t");
  show_value(privkey->d, 20);
  printf("\nkey modulus n =\n\t");
  show_value(pubkey->n, 20);
  printf("\n");

  while (1) {
    printf("Enter some data to be encrypted (ButWithNoSpaces).\nEnter a single period \".\" to quit.\n");
    scanf("%s", tmp);
    if (tmp[0] == '.')
      goto done;

    for (i = 0; i < MAX; i++)
      plaintext[i] = 0;

    /*  printf("string length = %d\n", strlen(tmp));*/
    for (i = 0; i < strlen(tmp); i++)
      plaintext[MAX - strlen(tmp) + i] = tmp[i];

    printf("Equivalent hexadecimal notation of your plaintext..\n");
    show_value(plaintext, 25);
    printf("\n");

    printf("RSA encrypted ciphertext is...\n");
    rsa_encrypt_bitwise(pubkey, plaintext, ciphertext);

    /*  printf("done encrypt\n");
    valcpy(ciphertext, tmp);
    while (tmp[0] == 0) {
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
    }
    printf("%s", tmp);*/

    show_value(ciphertext, 25);
    printf("\n");

    printf("RSA decrypted is...\n");
    rsa_decrypt_bitwise(privkey, ciphertext, plaintext);

    /*  valcpy(plaintext, tmp);
    while (tmp[0] == 0) {
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
      shift(tmp, LEFT);
    }
    printf("%s", tmp);*/

    show_value(plaintext, 25);
    printf("\n");
  }
 done:

  /*  add(p, q, big3);
    show_value(p, 10);
    printf(" + ");
    show_value(q, 10);
    printf(" = ");
    show_value(big3, 10);
    printf("\n");*/

  /*  sub(big1, big2, big3);
  printf("%d.%d.%d - %d.%d.%d = %d.%d.%d\n", big1[MAX - 3], big1[MAX - 2], big1[MAX - 1], big2[MAX - 3], big2[MAX - 2], big2[MAX - 1], big3[MAX - 3], big3[MAX - 2], big3[MAX - 1]);*/

  /*  expon(big4, big6, big3);
  show_value(big4, 10);
  printf(" ^ ");
  show_value(big6, 10);
  printf("\n = ");
  show_value(big3, 35);
  printf("\n");*/

  /*  printf("%d.%d.%d >> 1", big2[MAX - 3], big2[MAX - 2], big2[MAX - 1]);
  shift(big2, RIGHT);
  printf(" = %d.%d.%d\n", big2[MAX - 3], big2[MAX - 2], big2[MAX - 1]);*/
  /*  sqroot(big3, big5);*/

  /*
  printf("%d.%d.%d.%d.%d prime?  ==> %d\n", big6[MAX - 5], big6[MAX - 4], big6[MAX - 3], big6[MAX - 2], big6[MAX - 1], check_prime(big6));
  printf("%d.%d.%d.%d.%d prime?  ==> %d\n", big7[MAX - 5], big7[MAX - 4], big7[MAX - 3], big7[MAX - 2], big7[MAX - 1], check_prime(big7));

  find_prime(big3, big5);
  printf("lpgt( ");
  show_value(big3, 20);
  printf(" )\n = ");
  show_value(big5, 20);
  printf("\n");
  */

}
